"""Probability distribution objects."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
# adapted from sktime

__all__ = ["IID"]

from skpro.distributions.compose._iid import IID
