"""Module containing adapters other framework packages covering multiple tasks."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__all__ = ["_SksurvAdapter"]

from skpro.survival.adapters.sksurv import _SksurvAdapter
