"""Base module with base classes BaseObject, BaseEstimator."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__all__ = ["BaseEstimator", "BaseMetaEstimator", "BaseObject"]

from skpro.base._base import BaseEstimator, BaseMetaEstimator, BaseObject
