"""Reduction to probabilistic classification."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

from skpro.regression.binned._sklearn_bin_regressor import HistBinnedProbaRegressor

__all__ = ["HistBinnedProbaRegressor"]
