# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Module for nested equality checking."""
from skpro.utils.deep_equals._deep_equals import deep_equals

__all__ = [
    "deep_equals",
]
